
//
// runtime.h
//
// This is the description of the functions for the Cel runtime.
// 
//
// $Id: runtime.h,v 1.24 2001/02/04 08:51:59 dru Exp $
//
//
//
//


#include "svm32.h"
#include "cel.h"
#include "atom.h"
#include "memstream.h"

#ifndef H_CELRUNTIME
#define H_CELRUNTIME



#define nil   (void *)0
#define True  1
#define False 0




#define PROTO_ACTION_GET          0
#define PROTO_ACTION_SET          1
#define PROTO_ACTION_GETFRAME     2
#define PROTO_ACTION_SETFRAME     3
#define PROTO_ACTION_PASSTHRU     4
#define PROTO_ACTION_CODE         5
#define PROTO_ACTION_INTRINSIC    6




// G L O B A L S   &   H O O K S

// Fill factor for the proto hashes
extern int PROTO_FillFactor;

// Fill factor for the proto hashes
extern Proto Lobby;
extern Proto DefaultBehavior;
extern Proto Globals;
extern Proto Traits;
extern Proto Clonable;
extern Proto ObjectClass;
extern Proto StringParent;
extern Proto ArrayParent;
extern Proto OddBalls;

extern Proto IntegerClass;
extern Proto AtomClass;
extern Proto FloatClass;
extern Proto Capsule;

// Special Objects
extern Proto TrueObj;
extern Proto FalseObj;


extern int debugParse;
extern int debugParseLevel;
extern int debugDump;
extern int debugTrace;


// The default debugging info handler
extern int (*printString)(char * str);        /* Output handler */
extern int (*ioOut)(char * str, int count);   /* IO Output Handler */
extern int (*addPrintSpaces)(int32 spaces);   /* Output Handler */

int runtimeActivateTrace;


extern int runtimeActivateTrace;

extern VM32Cpu * Cpu;

extern Atom Parent;
extern Atom BlockParent;
extern Atom Code;
extern Atom DebugName;


int     celInit(int i, VM32Cpu * cpu);


void *  objectNew(int slotNum);
void *  objectNewWithSize(int slotNum, int dataNum);
void *  objectNewInteger(int i);
void *  objectNewString(char * str);
void *  objectNewBlob(void * blob);
// An object blob is what stores object references for hashes, or arrays
void *  objectNewObjectBlob(int num);
// What the size of a blob is
uint32 objectBlobSize (Proto obj);
void *  objectBlobResize (Proto obj, int num);
Proto   objectBlobAtGet(Proto obj, uint32 index);
void    objectBlobAtSet(Proto obj, uint32 index, Proto newItem);

void    objectFree(void * obj);


inline Proto   objectSetAsCodeProto(Proto proto);
inline Proto   objectSetAsBlockProto(Proto proto);
inline int     isObjectSetAsCodeProto(Proto proto);

inline void  * objectResolveForwarders(void * obj);


// Accessing objects -> these follow forwarders before accessing
// object
int32 objectIsType(Proto proto);
//void *  objectForKeyString(void * proto, char * str);
//void *  objectForKey(void * proto, char * key);
//void *  objectSlotAt(void * proto, int slotNum);
void *  objectSetSlot(Proto proto, Atom key, char action, uint32 data);

void *  objectSetDataAtIndex(Proto proto, uint32 index, uint32 data);

void *  objectEasySetSlot(Proto proto, Atom key, uint32 data);

void *  objectEasySetFrameSlot(Proto proto, Atom key, uint32 data);

void *  objectInsertSlot(Proto proto, uint32 key, uint32 action, uint32 index);

// Set another key/action to point to the same data as an existingKey
void *  objectSetSlotSame(Proto proto, Atom key, Atom existingKey, char action);

// Hash type lookup
inline int objectLookupKey(Proto proto, uint32 key, struct _protoEntry * * slotPtr);

// Accessing data from objects
//int     objectInteger(void * proto);
//void *  objectData(void * proto);

void * objectExpand(Proto proto);

// Enumerating through slots
ProtoSlot objectEnumerate(Proto proto, int * currentSlot, int skipParent);

// This gets the data from a slot, it returns true if it
// finds the key
//// Whether this function will exist is a good question
uint32 objectGetSlot(Proto proto, Atom key, Proto * result);

// This removes a slot from a proto 
// it returns true if it finds the key
uint32 objectDeleteSlot(Proto proto, Atom key);


// This returns a pointer to the data - this only works on blobs
void *  objectPointerValue(Proto proto);

uint32    objectIntegerValue(Proto proto);


inline int objectCapacity(Proto proto);
inline int objectSlotCount   (Proto proto);
inline int objectDataCount   (Proto proto);

// Dump the object to stdio
void *  objectDump(Proto proto);
void *  objectShallowDump(Proto proto);


// The magic function
// This function expects the proto, selector, and arguments on the vm stack
void activateSlot(void);


// locateTargetWithSlotVia
//    This is used by some intriniscs to lookup a slot via inheritance
//    since and intrinsic will sometimes have as a target the context
//    for example (at: 5) print. Will use the 'self' which will be
//    block or code context. I don't want to force people to do '(self at: 5) print.'
//    just yet. This is a sticky issue.
//

Proto locateObjectWithSlotVia (Proto key, Proto obj);


/////////////////////////////////////////////////////
//
// A C T I O N   H A N D L E R S
//
//


// Action Handlers - inlined because only activateSlot should call them
inline void actionGet(Proto object, ProtoSlot slotPtr);

// actionSet assumes that this slot was a 'SET' action. It assumes that the stack
// contained some destination object and the message activated this slot and the
// new object
//
inline void actionSet(Proto object, ProtoSlot slotPtr);


// actionGetStack assumes that this slot was a 'GET' action. It assumes that the stack
// contained some destination object and the message activated this slot
//
inline void actionGetStack(Proto object, ProtoSlot slotPtr, uint32 fp);


// actionSetStack locates the argument in the frame that has to be set and
// sets it to the argument passed in via the immediate stack. For example,
// imagine a local variable (which is stack mapped) in a code proto needs
// to get set. The trick is that the data for the set is in the immediate
// stack frame and the variable is relative to the fp
// now for the two lines of code...:)
//
inline void actionSetStack(Proto object, ProtoSlot slotPtr, uint32 fp);


// actionPassthru deals with activating some code for a passthru slot
//
// This will push the parent if 'useParamAsParent' is true otherwise
// it will retrieve the parent off of the stack. This is the target
// of the message that had the passThru slot.
inline void actionPassthru(Proto object, ProtoSlot slotPtr, Proto parent);


// actionCode - not implemented yet.
//
inline void actionCode(Proto object, ProtoSlot slotPtr);

// actionIntrinsic calls the code via C for the argument in the slot
//
inline void actionIntrinsic(Proto object, ProtoSlot slotPtr);


/////////////////////////////////////////////////////
//
// I N T R I N S I C S
//
//


void intrinsicSetSlotWith(void);

// Arguments: target 'traceOn/traceOff'
// Return: target
// Notes:
// This turns message tracing On or Off
void intrinsicTrace (void);

// Arguments: target 'isNil'
// Return: true/false
void intrinsicObjectIsNil(void);

// Arguments: target 'clone'
// Return: new clone
void intrinsicClone(void);

// Arguments: target <someMessage>
// Return: new clone
// Notes:
// Assumes some binary message, just returns the target, the self
void intrinsicSelf(void);

// Arguments: target '+' argument
// Return: newValue
// Notes:
// Simple add function - imagine the hardware to iplement this versus
// a simple adder circuit
void intrinsicIntegerAdd(void);


// Arguments: target '-' argument
// Return: newValue
// Notes:
// Simple integer subtract function
void intrinsicIntegerSubtract(void);

// Arguments: target '*' argument
// Return: newValue
// Notes:
// Simple integer multiply function
void intrinsicIntegerMultiply(void);

// Arguments: target '/' argument
// Return: newValue
// Notes:
// Simple integer divide function
void intrinsicIntegerDivide(void);

// Arguments: target 'abs'
// Return: newValue
// Notes:
// Simple integer absolute value function
void intrinsicIntegerAbsolute(void);

// Arguments: target '=' argument
// Return: true/false
// Notes:
// Simple integer equality function
void intrinsicIntegerEqual(void);

// Arguments: target '!=' argument
// Return: true/false
// Notes:
// Simple integer inequality function
void intrinsicIntegerNotEqual (void);

// Arguments: target '>' argument
// Return: true/false
// Notes:
// Simple integer greater-than function
void intrinsicIntegerGreater(void);

// Arguments: target '>=' argument
// Return: true/false
// Notes:
// Simple integer greater-than or equal function
void intrinsicIntegerGreaterEqual(void);

// Arguments: target '<' argument
// Return: true/false
// Notes:
// Simple integer less-than function
void intrinsicIntegerLess(void);

// Arguments: target '<=' argument
// Return: true/false
// Notes:
// Simple integer less-than or equal function
void intrinsicIntegerLessEqual(void);

// Arguments: target '@' argument
// Return: newValue
// Notes:
// Simple point create function 
void intrinsicPointCreate(void);

// Arguments: target 'print'
// Return: target
// Notes:
// Prints to the default string print routine
void intrinsicIntegerPrint(void);

// Arguments: target 'asString'
// Return: target
// Notes:
// This also handles 'asStringWithFormat' 
void intrinsicIntegerAsString(void);

// Arguments: target 'import:' argument
// Return: target
// Notes:
// Loads in that Kit
void intrinsicKitImport(void);

// A support method for KitImport
Proto intrinsicLoadModule(Atom fileName);

// Arguments: target 'importLib:' argument
// Return: target
// Notes:
// Loads in that Kit's library
void intrinsicKitImportLib (void);

// Arguments: target 'value'
// Return: NONE - the block does the return
// Notes:
// This changes the thread of control to a block
void intrinsicValue(void);

// Arguments: target 'whileTrue:' obj
// Return: NONE - the block does the return
// Notes:
// This changes the thread of control to a WhileTrue block
void intrinsicWhileTrue (void);

// Arguments: target 'whileFalse:' obj
// Return: NONE - the block does the return
// Notes:
// This changes the thread of control to a WhileFalse block
void intrinsicWhileFalse (void);


// Arguments: target 'print'
// Return: target
// Notes:
// Prints to the default string print routine
void intrinsicAtomPrint(void);


// Arguments: target '=' argument
// Return: true/false
// Notes:
// Simple atom equality function
void intrinsicAtomEqual(void);

// Arguments: target '>' argument
// Return: true/false
// Notes:
// Simple atom greater-than function
void intrinsicAtomGreater(void);

// Arguments: target '>=' argument
// Return: true/false
// Notes:
// Simple atom greater-than or equal function
void intrinsicAtomGreaterEqual(void);

// Arguments: target '<' argument
// Return: true/false
// Notes:
// Simple atom less-than function
void intrinsicAtomLess(void);

// Arguments: target '<=' argument
// Return: true/false
// Notes:
// Simple atom less-than or equal function
void intrinsicAtomLessEqual(void);

// Arguments: target 'size'
// Return: integer size
// Notes:
// Returns the size of the string
void intrinsicAtomSize(void);

// Arguments: target 'asString'
// Return: a new string
// Notes:
// Create a new String object out of this literal
void intrinsicAtomAsString(void);


// Arguments: target 'objectDump'
// Return: junk
void intrinsicObjectDump(void);

// Arguments: target 'objectShallowDump'
// Return: junk
void intrinsicObjectShallowDump(void);


// Arguments: target 'print'
// Return: self/target
// Notes:
// Prints to the default string print routine
void intrinsicStringPrint(void);

// Arguments: target '+' literal/string
// Return: new string that is the combination
// Notes:
// Creates a new string that is the concatenation of the target and the argument
void intrinsicStringAdd(void);

// Arguments: target 'insert:' literal/string 'at:' position
// Return: string that is the combination
// Notes:
// Returns the string with that insertion at that position
void intrinsicStringInsertAt(void);

// Arguments: target 'append:' literal/string
// Return: self
// Notes:
// Just appends a string on to the existing target
void intrinsicStringAppend(void);

// Arguments: target 'reverse'
// Return: self
// Notes:
// Reverses the string in place
void intrinsicStringReverse(void);

// Arguments: target 'lowercase'
// Return: self
// Notes:
// Makes each character lower case in the string
void intrinsicStringLowercase(void);

// Arguments: target 'setCapacity:' integer
// Return: new string that is of that capacity
// Notes:
// Creates a new string that is has that capacity
void intrinsicStringSetCapacity(void);


// Cel strstr - our own version that works on blobs rather than null-terminated
// sequences
char *celStrStr(char *haystack, int haySize,char *needle, int needleSize);

// Same as above but searches from last part of string
char *celRevStrStr(char *haystack, int haySize, char *needle, int needleSize);


// Arguments: target 'setCapacity:' integer
// Return: new string that is of that capacity
// Notes:
// This routine handles splitOn: and splitOn:atMostCount:
// count: 0 = stop at that number of splits is has that capacity
void intrinsicStringSplit(void);

// Creates a new string that is a copy of the specified arguments
Proto createString(char * buff, int size);

// Arguments: target 'copyFrom:' integer 'To:' integer
// Return: new string that is the slice of the original
// Notes:
// Currently it doesn't deal with negative offsets, but
// it will
void intrinsicStringSlice(void);

// Arguments: target 'asInteger'
// Return: new integer
// Notes:
// This routine tries to return the argument as an integer
// Otherwise it returns a nil
void intrinsicStringAsInteger(void);

// Arguments: target 'findLast:' argument
// Return: new integer
// Notes:
// This routine tries to find the argument from right to left
// Otherwise it returns a nil
void intrinsicStringFindLast(void);


// Arguments: target 'at:' integer
// Return: value in array
// Notes:
// Return the array position at: x
void intrinsicArrayAt(void);

// Arguments: target 'insert:' obj 'at:' integer 
// Return: self
// Notes:
// Insert an object at position
void intrinsicArrayInsertAt(void);

// Arguments: target 'at:otherwiseUse:' integer
// Return: value in array
// Notes:
// Return the array position at: x
void intrinsicArrayAtOtherwiseUse(void);

// Arguments: target 'at:' integer 'Put:' obj
// Return: self
// Notes:
// Place an object at position
void intrinsicArrayAtPut(void);

// Arguments: target 'removeAt:' integer
// Return: self
// Notes:
// Remove an object at position and slide the rest to the left
void intrinsicArrayRemoveAt(void);

// Arguments: target 'append:' object
// Return: self
// Notes:
// Append an object to the end of the array
void intrinsicArrayAppend(void);

// Arguments: target 'getFromFile:' argument
// Return: new structure or nil
// Notes:
// Loads in that simpledata
void intrinsicSDGetFromFile (void);

// Arguments: target 'saveToFile:' argument
// Return: target
// Notes:
// 
void intrinsicSDSaveToFile (void);



// U S E R  S U P P O R T functions


VM32Cpu * doCelInit (char * module, int32 noRun);

Proto loadModule(char * fileName);


int  checkCompiled(char * fileName);

MemStream loadModuleStream(char * fileName);

int  writeModule(char * fileName, void * mem, int length);


#endif
